/**
 * Copyright (c) 2013 Perto S.A. All rights reserved.
 *
 * \file pertousb.c
 *
 * \brief Backend CUPS component for Perto Printer TEC.
 *
 * Implements the low-level communication with the printer.
 *
 * \author Abilio Faria
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>

#include "usbprinter.h"

/*!
 * Private variables
 */
volatile int canceled = 0;

/**
 * \brief Cancel the job
 * \param sig signal 
 */ 
void cancel_job(int sig)
{
    (void)sig;

    fprintf(stderr, "ERROR: pertousb: Job canceled\n");
    canceled = 1;
}

/**
 * \brief main
 * Command line usage: pertousb job user title copies options [filename]
 * \param argc argc 
 * \param argv argv[]
 * \return 0 (success), 1 (failure)
 */
int main(int argc, char *argv[])
{
    usb_printer printer;
    FILE *file;
    int i;
    char buf[32];

#if defined(HAVE_SIGACTION) && !defined(HAVE_SIGSET)
    struct sigaction action;          /*!< POSIX signals actions */
#endif /* HAVE_SIGACTION && !HAVE_SIGSET */

    if (argc == 1) {
        /* this printf is used by CUPS to identify the printer. Do not change this */ 
        printf("network pertousb:/tmp/printer1 \"Unknown\" \"Perto Printer TEC (libusb)\"\n");
        return 1;
    }

    if (argc < 6 || argc > 7) {
        fprintf(stderr, "ERROR: Usage: %s job user title copies options [filename]\n", argv[0]);
        return 1;
    }

#ifdef HAVE_SIGSET /* Usa sinais System V para POSIX (evita bugs) */
    sigset(SIGTERM, cancel_job);
#elif defined(HAVE_SIGACTION)
    memset(&action, 0, sizeof(action));
    sigemptyset(&action.sa_mask);
    action.sa_handler = cancel_job;
    sigaction(SIGTERM, &action, NULL);
#else
    signal(SIGTERM, cancel_job);
#endif /* HAVE_SIGSET */

    if (argc == 6) {
        file = stdin;
        fprintf(stderr, "ERROR: pertousb: Using stdin\n");
    } else {
        file = fopen(argv[6], "rb");
        if (file == NULL) {
            fprintf(stderr, "ERROR: pertousb: Failed to open file: %s\n", strerror(errno));
            return 1;
        }
        fprintf(stderr, "ERROR: pertousb: Using file: %s\n", argv[6]);
    }

    printer = usb_printer_open();
    if (printer == NULL) {
        fprintf(stderr, "ERROR: pertousb: Failed to open printer: %s\n", usb_strerror());
        return 1;
    } 

    fprintf(stderr, "DEBUG: pertousb: Start file printing\n");

    while(!feof(file) || !ferror(file) || !canceled) {
        i = fread(buf, 1, 32, file);
        if (i <= 0) {
            break;
        }    
        if (usb_printer_write(printer, buf, i) < 1) {
            fprintf(stderr, "ERROR: pertousb: Timeout: %s\n", usb_strerror());
            usb_printer_close(printer);
            return 1;
        }
    }

    fprintf(stderr, "ERROR: pertousb: Ready to print\n");

    usb_printer_close(printer);
    return canceled;
}
