/**
 * Copyright (c) 2013 Perto S.A. All rights reserved.
 *
 * \file rastertoperto.c
 *
 * \brief Printer driver filter that convert CUPS raster files into a format usable by Perto Printer TEC.
 *
 * \author Abilio Faria
 */

#include <cups/raster.h>
#include <cups/ppd.h>
#include <fcntl.h>
#include <signal.h>

/*!
 * Private variables
 */
volatile int canceled;   /*!< Non-zero if job is canceled */

/*!
 * Prototypes
 */
void cancel_job(int sig);
void register_cancel_job( void);
void deregister_cancel_job( void);
void printer_cmd_set_line_spacing(void);
void printer_cmd_CR(void);
void printer_cmd_LF(void);
void printer_cmd_raster_bit_image(unsigned int width, unsigned int height);
void printer_cmd_eject_page(void);
void printer_cmd_cut_paper(void);
void cut_paper_page(cups_page_header2_t *header);
void cut_paper_job(cups_page_header2_t *header);
int get_non_blank_lines(unsigned char *buffer, unsigned int height, unsigned int bytes_per_line);
void start_page(ppd_file_t *ppd, cups_page_header2_t *header);
void end_page(ppd_file_t *ppd, cups_page_header2_t *header);

/**
 * \brief Cancel the job
 * \param sig signal 
 */ 
void cancel_job(int sig)
{
    (void)sig;

    fprintf(stderr, "ERROR: rastertoperto: Job canceled\n");
    canceled = 1;
}

/**
 * \brief Register the SIGTERM cancel_job
 * Register the signal handler to treat the job canceling.
 * \param sig signal 
 */ 
void register_cancel_job( void)
{
#if defined(HAVE_SIGACTION) && !defined(HAVE_SIGSET)
    struct sigaction action;          /*!< POSIX signals actions */
#endif /* HAVE_SIGACTION && !HAVE_SIGSET */

#ifdef HAVE_SIGSET
    sigset(SIGTERM, cancel_job);
#elif defined(HAVE_SIGACTION)
    memset(&action, 0, sizeof(action));
    sigemptyset(&action.sa_mask);
    action.sa_handler = cancel_job;
    sigaction(SIGTERM, &action, NULL);
#else
    signal(SIGTERM, cancel_job);
#endif /* HAVE_SIGSET */
}

/**
 * \brief Deregister the SIGTERM cancel_job
 * Deregister the signal handler to treat the job canceling.
 * \param sig signal 
 */ 
void deregister_cancel_job( void)
{
#if defined(HAVE_SIGACTION) && !defined(HAVE_SIGSET)
    struct sigaction action;          /*!< POSIX signals actions */
#endif /* HAVE_SIGACTION && !HAVE_SIGSET */

#ifdef HAVE_SIGSET
    sigset(SIGTERM, SIG_IGN);
#elif defined(HAVE_SIGACTION)
    memset(&action, 0, sizeof(action));
    sigemptyset(&action.sa_mask);
    action.sa_handler = SIG_IGN;
    sigaction(SIGTERM, &action, NULL);
#else
    signal(SIGTERM, SIG_IGN);
#endif /* HAVE_SIGSET */
}

/**
 * \brief Printer command to set line spacing to 0
 */
void printer_cmd_set_line_spacing(void)
{
    unsigned char cmd[3] = {0x1B, '3', 0x00};   /* ESC 3 00H */    
    fwrite(cmd, 1, 3, stdout);   
}

/**
 * \brief Printer command to print a CR
 * Set absolute print position to 0
 */
void printer_cmd_CR(void)
{
    unsigned char cmd[4] = {0x1B, '$', 0x00, 0x00};   /* ESC $ 00H 00H */    
    fwrite(cmd, 1, 4, stdout);   
}

/**
 * \brief Printer command to print a  LF
 * Prints the data in the print buffer and feeds one line, based on the current line spacing.
 */
void printer_cmd_LF(void)
{
    unsigned char cmd[1] = {0x0A};   /* LF */    
    fwrite(cmd, 1, 1, stdout);   
}

/**
 * \brief Printer command to print a raster bit image
 * \param width width in bits
 * \param height height in bits
 */
void printer_cmd_raster_bit_image(unsigned int width, unsigned int height)
{
    unsigned char cmd[8] = {0x1D, 'v', '0', 0x00, 0x00, 0x00, 0x00, 0x00};   /* GS v 0 00H xl xh yl yh */    
    cmd[4] = width & 0xFF;
    cmd[5] = (width >> 8) & 0xFF;    
    cmd[6] = height & 0xFF;
    cmd[7] = (height >> 8) & 0xFF;    
    fwrite(cmd, 1, 8, stdout);   
}

/**
 * \brief Printer command to eject page without cut paper
 */
void printer_cmd_eject_page(void)
{
    printer_cmd_LF();   /* LF */
}

/**
 * \brief Printer command to cut paper
 */
void printer_cmd_cut_paper(void)
{
    unsigned char cmd[3] = {0x1D, 'V', 0x01};   /* GS V 01H */
    fwrite(cmd, 1, 3, stdout);
}

/**
 * \brief Cut paper in the end of page if needed
 * \param header page header
 */ 
void cut_paper_page(cups_page_header2_t *header)
{
    if (header->CutMedia == CUPS_CUT_NONE) {
        /* eject page */
        printer_cmd_eject_page();
        fprintf(stderr, "ERROR: rastertoperto: Eject page\n");
    }
    else if (header->CutMedia == CUPS_CUT_PAGE) {
        /* cut paper */
        printer_cmd_cut_paper();
        fprintf(stderr, "ERROR: rastertoperto: Cut paper\n");
    }
}

/**
 * \brief Cut paper in the end of job if needed
 * \param header page header
 */ 
void cut_paper_job(cups_page_header2_t *header)
{
    if (header->CutMedia == CUPS_CUT_JOB) {
        /* cut paper */
        printer_cmd_cut_paper();
        fprintf(stderr, "ERROR: rastertoperto: Cut paper\n");
    }
}

/**
 * \brief Get number of lines after removing the blank lines in the end of the page
 * \param buffer buffer
 * \param height height in bits
 * \param bytes_per_line bytes per line
 * \return number of lines after removing the blank lines in the end of the page
 */
int get_non_blank_lines(unsigned char *buffer, unsigned int height, unsigned int bytes_per_line)
{
    int last_line;
    int x;
    unsigned char *pbuffer;

    for (last_line = height - 1; last_line >= 0; last_line--) {
        pbuffer = buffer + (last_line * bytes_per_line);
        for (x = 0; x < bytes_per_line; x++) {
            if (*pbuffer != 0) {
                return (last_line + 1);
            }
            pbuffer++;
        }
    }
    return 0;
}
 
/**
 * \brief Start page processing
 * \param ppd ppd
 * \param header header
 */ 
void start_page(ppd_file_t *ppd, cups_page_header2_t *header)
{
    fprintf(stderr, "ERROR: rastertoperto: Start page\n");

    /* print the header fields */
    fprintf(stderr, "DEBUG: MediaClass = \"%s\"\n", header->MediaClass);
    fprintf(stderr, "DEBUG: MediaColor = \"%s\"\n", header->MediaColor);
    fprintf(stderr, "DEBUG: MediaType = \"%s\"\n", header->MediaType);
    fprintf(stderr, "ERROR: OutputType = \"%s\"\n", header->OutputType);
    fprintf(stderr, "DEBUG: AdvanceDistance = %d\n", header->AdvanceDistance);
    fprintf(stderr, "DEBUG: AdvanceMedia = %d\n", header->AdvanceMedia);
    fprintf(stderr, "DEBUG: Collate = %d\n", header->Collate);
    fprintf(stderr, "ERROR: CutMedia = %d\n", header->CutMedia);
    fprintf(stderr, "DEBUG: Duplex = %d\n", header->Duplex);
    fprintf(stderr, "ERROR: HWResolution = [ %d %d ]\n", header->HWResolution[0], header->HWResolution[1]);
    fprintf(stderr, "ERROR: ImagingBoundingBox = [ %d %d %d %d ]\n",
            header->ImagingBoundingBox[0], header->ImagingBoundingBox[1],
            header->ImagingBoundingBox[2], header->ImagingBoundingBox[3]);
    fprintf(stderr, "ERROR: InsertSheet = %d\n", header->InsertSheet);
    fprintf(stderr, "DEBUG: Jog = %d\n", header->Jog);
    fprintf(stderr, "ERROR: LeadingEdge = %d\n", header->LeadingEdge);
    fprintf(stderr, "ERROR: Margins = [ %d %d ]\n", header->Margins[0], header->Margins[1]);
    fprintf(stderr, "ERROR: ManualFeed = %d\n", header->ManualFeed);
    fprintf(stderr, "ERROR: MediaPosition = %d\n", header->MediaPosition);
    fprintf(stderr, "DEBUG: MediaWeight = %d\n", header->MediaWeight);
    fprintf(stderr, "DEBUG: MirrorPrint = %d\n", header->MirrorPrint);
    fprintf(stderr, "DEBUG: NegativePrint = %d\n", header->NegativePrint);
    fprintf(stderr, "DEBUG: NumCopies = %d\n", header->NumCopies);
    fprintf(stderr, "ERROR: Orientation = %d\n", header->Orientation);
    fprintf(stderr, "DEBUG: OutputFaceUp = %d\n", header->OutputFaceUp);
    fprintf(stderr, "ERROR: PageSize = [ %d %d ]\n", header->PageSize[0], header->PageSize[1]);
    fprintf(stderr, "DEBUG: Separations = %d\n", header->Separations);
    fprintf(stderr, "DEBUG: TraySwitch = %d\n", header->TraySwitch);
    fprintf(stderr, "DEBUG: Tumble = %d\n", header->Tumble);
    fprintf(stderr, "ERROR: cupsWidth = %d\n", header->cupsWidth);
    fprintf(stderr, "ERROR: cupsHeight = %d\n", header->cupsHeight);
    fprintf(stderr, "ERROR: cupsMediaType = %d\n", header->cupsMediaType);
    fprintf(stderr, "ERROR: cupsBitsPerColor = %d\n", header->cupsBitsPerColor);
    fprintf(stderr, "ERROR: cupsBitsPerPixel = %d\n", header->cupsBitsPerPixel);
    fprintf(stderr, "ERROR: cupsBytesPerLine = %d\n", header->cupsBytesPerLine);
    fprintf(stderr, "ERROR: cupsColorOrder = %d\n", header->cupsColorOrder);
    fprintf(stderr, "ERROR: cupsColorSpace = %d\n", header->cupsColorSpace);
    fprintf(stderr, "ERROR: cupsCompression = %d\n", header->cupsCompression);

    /* set line spacing */
    printer_cmd_set_line_spacing();
}

/**
 * \brief End page processing
 * \param ppd ppd
 * \param header header
 */
void end_page(ppd_file_t *ppd, cups_page_header2_t *header)
{
    fprintf(stderr, "ERROR: rastertoperto: End page\n");

    /* cut the paper at the end of the page if needed */
    cut_paper_page(header);
}
 
/**
 * \brief main
 * \param argc argc 
 * \param argv argv[]
 * \return 0 (success), 1 (failure)
 */
int    main(int argc, char *argv[])
{
    cups_raster_t *ras;               /*!< Raster stream for printing */
    cups_page_header2_t header;       /*!< Page header from file */
    int fd;                           /*!< File descriptor */
    ppd_file_t *ppd;                  /*!< PPD file */
    int num_options = 0;              /*!< Number of CUPS options */
    cups_option_t *options = NULL;    /*!< CUPS options */
    int page = 0;                     /*!< Page counter */
    int lines;                        /*!< Lines counter */
    unsigned char *buffer;            /*!< pointer to buffer to read the lines */
    unsigned char *pbuffer;           /*!< pointer to the beginning of a line in the buffer */
    int remove_blank = 0;             /*!< remove blank at the end of page */
    ppd_option_t *remove_blank_option;       /*!< remove blank option */
    ppd_choice_t *remove_blank_choice_on;    /*!< remove blank choice On */

    /* make sure status messages are not buffered */
    setbuf(stderr, NULL);

    /* check command-line */
    if (argc < 6 || argc > 7) {
        /* we don't have the correct number of arguments; write an error message and return */
        fprintf(stderr, "ERROR: rastertoperto job-id user title copies options [file]\n");
        return 1;
    }

    /* open the page stream */
    if (argc == 7) {
        if ((fd = open(argv[6], O_RDONLY)) == -1) {
            fprintf(stderr, "ERROR: rastertoperto: Unable to open raster file");
            return 1;
        }
    }
    else {
        fd = 0;
    }
    ras = cupsRasterOpen(fd, CUPS_RASTER_READ); 

    /* open the PPD file and apply options */
    num_options = cupsParseOptions(argv[5], 0, &options);
    ppd = ppdOpenFile(getenv("PPD"));
    if (ppd != NULL) {
        ppdMarkDefaults(ppd);
        /* mark command-line options in a PPD file */
        cupsMarkOptions(ppd, num_options, options);
    }
    /* free the options */
    cupsFreeOptions(num_options, options);

    /* set if the blank in the end of the page should be removed */
    remove_blank_option = ppdFindOption(ppd, "RemoveBlank");
    remove_blank_choice_on  = ppdFindChoice(remove_blank_option, "On");
    remove_blank = (remove_blank_choice_on->marked == 1);
    fprintf(stderr, "ERROR: rastertoperto: Remove blank (End Of Page) %s\n", ((remove_blank == 0)?"Off":"On"));

    /* process each page */
    canceled = 0;
    while (cupsRasterReadHeader2(ras, &header)) {
    
        if (header.cupsWidth < 1 ) {
            fprintf(stderr, "ERROR: rastertoperto: Failed. Invalid header.cupsWidth");
            break;
        }    

        /* allocate memory buffer for header.cupsHeight lines */
        buffer = malloc(header.cupsBytesPerLine * header.cupsHeight);
        if (buffer == NULL) {
            fprintf(stderr, "ERROR: rastertoperto: Failed. Not enough memory");
            break;
        }
        /* initialize the buffer */
        memset(buffer, 0, (header.cupsBytesPerLine * header.cupsHeight));

        /* register SIGTERM */
        register_cancel_job();

        /* setup this page */
        page ++;
        fprintf(stderr, "PAGE: %d %d\n", page, header.NumCopies);
        start_page(ppd, &header);

        /* read raster data */
        pbuffer = buffer;
        for (lines = 0; lines < header.cupsHeight; lines++) {

            /* read one raster bit image line to the buffer */
            if (cupsRasterReadPixels(ras, pbuffer, header.cupsBytesPerLine) == 0) {
                fprintf(stderr, "ERROR: rastertoperto: Failed. cupsRasterReadPixels() returns 0");
                break;
            }    
            /* advance one line in the buffer */
            pbuffer += header.cupsBytesPerLine;
        }

        /* check if the blank in the end of the page should be removed */
        if (remove_blank) {
            /* get number of lines after removing the blank lines in the end of the page */
            lines = get_non_blank_lines(buffer, lines, header.cupsBytesPerLine);
        }

        /* print 'line' lines at once */
        if (lines > 0) {
            printer_cmd_raster_bit_image(header.cupsBytesPerLine, lines);
            fwrite(buffer, 1, (header.cupsBytesPerLine * lines), stdout);   
        }

        /* finish this page */
        end_page(ppd, &header);

        /* release buffer */
        free(buffer);

        /* deregister SIGTERM */
        deregister_cancel_job();

        if (canceled) {
            break;
        }
    }
    /* end of the document */

    /* cut the paper at the end of the job if needed */
    cut_paper_job(&header);

    /* close the raster stream */
    cupsRasterClose(ras);
    if (fd != 0) {
        close(fd);
    }

    /* close the PPD file */
    if (ppd != NULL) {
        ppdClose(ppd);
    }

    fprintf(stderr, "ERROR: rastertoperto: Finished. Ready to print.\n");

    return canceled;
}
