/**
 * Copyright (c) 2013 Perto S.A. All rights reserved.
 *
 * \file usbprinter.c
 *
 * \brief libusb shell for USB port access used in Perto Printer TEC CUPS driver. 
 *
 * \author Abilio Faria
 */

#include "usbprinter.h"

/**
 * \brief Open USB device
 * \return Pointer to USB device handle (success) or NULL (failure)
 */
usb_printer usb_printer_open( void)
{
    struct usb_bus *busses, *bus;
    struct usb_device *printer_dev = NULL;
    struct usb_dev_handle *printer = NULL;

    usb_init();
    if (usb_find_busses() < 0) {
        return NULL;
    }
    if (usb_find_devices() < 0) {
        return NULL;
    }

    busses = usb_get_busses();
    for(bus = busses; bus; bus = bus->next) {
        struct usb_device *dev;
        for (dev = bus->devices; dev; dev = dev->next) {
            if (dev->descriptor.idVendor == USBPRINTER_VID && 
                    dev->descriptor.idProduct == USBPRINTER_PID) {
                printer_dev = dev;
                break;
            }
        }
    }
    if (printer_dev == NULL) {
        return NULL;
    }

    printer = usb_open(printer_dev);
    if (printer == NULL) {
        return NULL;
    }

    if ( usb_claim_interface(printer, 0) != 0 ) {
        usb_close(printer);
        return NULL;
    }

    return printer;
}

/**
 * \brief Close USB device
 * \param printer Pointer to USB device handle
 */
void usb_printer_close(usb_printer printer)
{
    if (printer == NULL)
        return;

    usb_release_interface(printer, 0);
    usb_close(printer);
    printer = NULL;
}

/**
 * \brief Write data into USB device
 * \param printer Pointer to USB device handle
 * \param buffer Pointer to buffer with data to be written
 * \param length number of bytes to be written
 * \return > 0 (success), <= 0 (failure)
 */
int usb_printer_write(usb_printer printer, char *buffer, int length)
{
    if (printer == NULL)
        return -1;

    return usb_bulk_write(printer, USBPRINTER_EP_WRITE, buffer, length, USBPRINTER_WRITE_TIMEOUT);
}
